{ File   : MakDH2K5.dpr
  Author : Deniz Oezmen
  Created: 2009-08-03
  Changed: 2009-08-15

  Deer Hunter APK creator (experimental), based on Luigi Auriemma's
  deer_hunter_2005.bms. Thus:


  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
}
program MakDH2K5;

{$APPTYPE CONSOLE}

uses
  SysUtils, SCRC, ZLib;

const
  MasterKeyPhrase: string = 'dh2005.exe;d3d.dll';

  APKEntryEncrypted  = 1;
  APKEntryCompressed = 2;

type
  TAPKHeader = packed record
    ID      : array[0..2] of Char;
    Ver     : Byte;
    DirOfs  : Longword;
    DirSize : Longword;
    AddRound: Boolean;
  end;

  TAPKEntry = packed record
    Name   : string;
    Ofs    : Longword;
    CSize  : Longword;
    Flags  : Byte;
    USize  : Longword;
    CRC    : Longword;
    Unknown: Longword;
  end;

  TAPKEntries = array of TAPKEntry;

  TBuffer = array of Byte;

var
  MasterKey : Longword;
  CRC       : TCRC;
  FileSpec  : AnsiString = '';
  DirSpec   : AnsiString = '';
  DoCompress: Boolean = False;
  DoEncrypt : Boolean = False;

procedure MasterKeyInit;
var
  i,
  Temp: Longword;
begin
  MasterKey := 1;

  for i := 1 to Length(MasterKeyPhrase) do
  begin
    Temp := (MasterKey and $0000ffff + Ord(MasterKeyPhrase[i])) mod $0000fff1;
    asm
      push ebx;
      mov  eax, MasterKey;
      sar  eax, 16;
      mov  ebx, Temp;
      add  eax, ebx;
      mov  ecx, $0000fff1;
      cdq;
      idiv ecx;
      shl  edx, 16;
      add  edx, ebx;
      mov  MasterKey, edx;
      pop  ebx;
    end
  end;
end;

procedure Encrypt(var Buffer: TBuffer; Key: Longword;
  AdditionalKeyRound: Boolean);
var
  i: Longint;
begin
  if AdditionalKeyRound then
  asm
    mov eax, Key;
    mov ecx, Key;
    and eax, $0000ffff;
    mov edx, $00006054;
    mul edx;
    sar ecx, 16;
    add eax, ecx;
    mov Key, eax;
  end;

  for i := 0 to High(Buffer) do
  begin
    asm
      mov eax, Key;
      mov ecx, Key;
      and eax, $0000ffff;
      mov edx, $00006054;
      mul edx;
      sar ecx, 16;
      add eax, ecx;
      mov Key, eax;
    end;
    Buffer[i] := Buffer[i] xor Key
  end;
end;

function Compress(var Buffer: TBuffer; Force: Boolean = False): Boolean;
var
  CompSize  : Longint;
  TempBuffer: Pointer;
begin
  CompressBuf(@Buffer[0], Length(Buffer), TempBuffer, CompSize);

  Result := (CompSize < Length(Buffer)) or Force;
  if Result then
  begin
    SetLength(Buffer, CompSize);
    Move(TempBuffer^, Buffer[0], CompSize);
  end;

  FreeMem(TempBuffer, CompSize)
end;

procedure ScanDir(const Dir, RelDir: string; var APKFile: file;
  const APKHeader: TAPKHeader; var APKEntries: TAPKEntries);
var
  Idx      : Longint;
  SearchRec: TSearchRec;
  InFile   : file;
  Buf      : TBuffer;
begin
  if FindFirst(Format('%s*.*', [Dir]), faAnyFile, SearchRec) = 0 then
    repeat
      with SearchRec do
      begin
        if (SearchRec.Attr and faDirectory) <> 0 then
        begin
          if (Name = '.') or (Name = '..') then
            Continue;

          // recursion for subdirectories
          ScanDir(Format('%s%s\', [Dir, Name]), Format('%s%s\', [RelDir, Name]),
            APKFile, APKHeader, APKEntries)
        end
        else
        begin
          WriteLn(Format('%s%s', [RelDir, SearchRec.Name]));

          // copy file data
          FileMode := fmOpenRead;
          AssignFile(InFile, Format('%s%s', [Dir, Name]));
          {$I-}
          Reset(InFile, 1);
          {$I+}
          if IOResult <> 0 then
            WriteLn(Format('Warning: Could not add "%s%s"', [Dir, Name]))
          else
          begin
            // prepare directory for file entry
            SetLength(APKEntries, Length(APKEntries) + 1);
            Idx := High(APKEntries);
            with APKEntries[Idx] do
            begin
              Name := Format('%s%s', [RelDir, SearchRec.Name]);
              Ofs := FilePos(APKFile);
              Flags := 0;
              Unknown := 0
            end;

            // read data from input file
            SetLength(Buf, Size);
            BlockRead(InFile, Buf[0], Length(Buf));
            APKEntries[Idx].USize := Length(Buf);

            // calculate checksum
            CRC.Init;
            CRC.Update(Buf[0], Length(Buf));
            CRC.Final(APKEntries[Idx].CRC);

            // try and compress data
            if DoCompress then
              if Compress(Buf) then
                APKEntries[Idx].Flags := APKEntries[Idx].Flags or
                  APKEntryCompressed;
            APKEntries[Idx].CSize := Length(Buf);

            // encrypt data
            if DoEncrypt then
            begin
              with APKEntries[Idx] do
                Encrypt(Buf, MasterKey xor USize xor Ofs xor CRC,
                  APKHeader.AddRound);
              APKEntries[Idx].Flags := APKEntries[Idx].Flags or
                APKEntryEncrypted
            end;

            // write data to archive
            BlockWrite(APKFile, Buf[0], Length(Buf));

            // clean up
            SetLength(Buf, 0);
            CloseFile(InFile)
          end
        end
      end
    until FindNext(SearchRec) <> 0;

  FindClose(SearchRec)
end;

procedure MakeDirTable(const APKEntries: TAPKEntries; out DirBuffer: TBuffer);
var
  i,
  Idx : Longint;
begin
  Idx := 0;
  SetLength(DirBuffer, 0);

  for i := 0 to High(APKEntries) do
    with APKEntries[i] do
    begin
      SetLength(DirBuffer, Length(DirBuffer) + Length(Name) + 1);
      Move(Name[1], DirBuffer[Idx], Length(Name));
      Inc(Idx, Length(Name));
      DirBuffer[Idx] := 0;
      Inc(Idx);
      SetLength(DirBuffer, Length(DirBuffer) + SizeOf(Ofs));
      Move(Ofs, DirBuffer[Idx], SizeOf(Ofs));
      Inc(Idx, SizeOf(Ofs));
      SetLength(DirBuffer, Length(DirBuffer) + SizeOf(CSize));
      Move(CSize, DirBuffer[Idx], SizeOf(CSize));
      Inc(Idx, SizeOf(CSize));
      SetLength(DirBuffer, Length(DirBuffer) + SizeOf(Flags));
      Move(Flags, DirBuffer[Idx], SizeOf(Flags));
      Inc(Idx, SizeOf(Flags));
      SetLength(DirBuffer, Length(DirBuffer) + SizeOf(USize));
      Move(USize, DirBuffer[Idx], SizeOf(USize));
      Inc(Idx, SizeOf(USize));
      SetLength(DirBuffer, Length(DirBuffer) + SizeOf(CRC));
      Move(CRC, DirBuffer[Idx], SizeOf(CRC));
      Inc(Idx, SizeOf(CRC));
      SetLength(DirBuffer, Length(DirBuffer) + SizeOf(Unknown));
      Move(Unknown, DirBuffer[Idx], SizeOf(Unknown));
      Inc(Idx, SizeOf(Unknown))
    end
end;

procedure MakeAPK(const APKName, DirName: string);
var
  APKFile   : file;
  APKHeader : TAPKHeader;
  APKEntries: TAPKEntries;
  DirBuffer : TBuffer;
begin
  // set up encryption scheme
  MasterKeyInit;

  // check for source directory
  if not DirectoryExists(DirName) then
  begin
    WriteLn(Format('could not access source directory "%s"', [DirName]));
    Exit;
  end;

  // open target archive
  AssignFile(APKFile, APKName);
  {$I-}
  Rewrite(APKFile, 1);
  {$I+}
  if IOResult <> 0 then
  begin
    WriteLn(Format('could not create target archive "%s"', [APKName]));
    Exit
  end;

  // generate and write preliminary header
  with APKHeader do
  begin
    ID := 'APK';
    Ver := 3;
    DirOfs := 0; // to be filled later
    DirSize := 0; // to be filled later
    AddRound := True
  end;
  BlockWrite(APKFile, APKHeader, SizeOf(APKHeader));

  // traverse source directory and copy file data
  CRC := TCRC.Create;
  SetLength(APKEntries, 0);
  ScanDir(IncludeTrailingPathDelimiter(ExpandFileName(DirName)), '',
    APKFile, APKHeader, APKEntries);
  CRC.Free;

  // create directory table
  MakeDirTable(APKEntries, DirBuffer);
  APKHeader.DirOfs := FilePos(APKFile);
  APKHeader.DirSize := Length(DirBuffer);
  Compress(DirBuffer, True);
  Encrypt(DirBuffer, MasterKey xor APKHeader.DirOfs xor APKHeader.DirSize,
    APKHeader.AddRound);
  BlockWrite(APKFile, DirBuffer[0], Length(DirBuffer));

  // write completed header
  Seek(APKFile, 0);
  BlockWrite(APKFile, APKHeader, SizeOf(APKHeader));

  CloseFile(APKFile)
end;

function ProcessParams: Boolean;
var
  i: Longint;
begin
  Result := False;

  if ParamCount = 0 then
    Exit;

  for i := 1 to ParamCount do
  begin
    if AnsiLowerCase(ParamStr(i)) = '/compress' then
      DoCompress := True
    else if AnsiLowerCase(ParamStr(i)) = '/encrypt' then
      DoEncrypt := True
    else if ParamStr(i)[1] = '/' then
      Exit
    else if FileSpec = '' then
      FileSpec := ParamStr(i)
    else if DirSpec <> '' then
      Exit
    else
      DirSpec := ParamStr(i)
  end;

  Result := (FileSpec <>  '') and (DirSpec <> '')
end;

begin
  if not ProcessParams then
    WriteLn('incorrect number of parameters')
  else
    MakeAPK(FileSpec, DirSpec)
end.
